TYPE = "vulnerability"

NAME = "Fingerprint web application framework"
SHORT_NAME = NAME

WSTG_CODE = ["WSTG-INFO-08"]

DESCRIPTION = (
    "The version of a web application framework can be identified due to the presence of its specific fingerprints."
)

SOLUTION = "This is only for informational purposes."

REFERENCES = [
    {
        "title": "OWASP: Fingerprint Web Application Framework",
        "url": (
            "https://owasp.org/www-project-web-security-testing-guide/stable/4-Web_Application_Security_Testing/"
            "01-Information_Gathering/08-Fingerprint_Web_Application_Framework.html"
        )
    }
]
